# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor Invoice batch handler for LOC SMS
"""

from __future__ import unicode_literals

import os
import logging
from decimal import Decimal

from locsms.db import model as locsms
from locsms.db import Session

from rattail.db.batch.vendorinvoice import VendorInvoiceRow
from rattail.db.batch.vendorinvoice.handler import VendorInvoiceHandler as BaseHandler
from rattail.db.api import get_product_by_upc
from rattail_locsms.sil import SMSWriter
from rattail.gpc import GPC
from rattail_locsms.deploy import deploy_to_outbox


log = logging.getLogger(__name__)


class VendorInvoiceHandler(BaseHandler):
    """
    Custom vendor invoice batch handler for use with LOC SMS.  This handler is
    able to reconcile the invoice against an existing purchasing document
    within SMS.  When the handler executes the batch it will generate SIL for
    updating the SMS purchasing document.
    """
    po_number_title = "SMS Trans. No."

    def get_purchase_order(self, number):
        """
        Fetch the purchasing document from SMS.
        """
        session = Session()
        purchase = session.query(locsms.Purchase).get(int(number))
        if purchase:
            purchase.details    # no-op to preload the relationship
        session.close()
        return purchase

    def validate_po_number(self, number, vendor):
        """
        If the given purchase document specified by ``number`` is for the
        correct vendor and is closed in the purchasing phase, this method does
        nothing interesting.  Any test the PO doesn't pass will cause an
        exception with a relevant message.
        """
        purchase = self.get_purchase_order(number)
        if not purchase:
            raise ValueError("SMS document not found for transaction "
                             "number: {0}".format(repr(number)))
        if purchase.F27 != vendor.id:
            raise ValueError("SMS transaction {0} is for the wrong vendor: "
                             "{1}".format(number, purchase.F334))
        if purchase.F1068 != 'ORDER':
            raise ValueError("SMS transaction {0} should have 'ORDER' transaction mode "
                             "but instead has: {1}".format(number, repr(purchase.F1068)))
        if purchase.F1067 != 'CLOSE':
            raise ValueError("SMS transaction {0} should have 'CLOSE' registration mode "
                             "but instead has: {1}".format(number, repr(purchase.F1067)))

    def cognize_row(self, session, row):
        """
        Round all case costs to 2 decimal places, since that is what will be in
        the SMS purchase document.  This helps to avoid false positives when
        checking for differences between the PO and invoice.
        """
        super(VendorInvoiceHandler, self).cognize_row(session, row)
        if row.case_cost:
            row.case_cost = row.case_cost.quantize(Decimal('0.01'))

    def cognize_purchase_order(self, session, invoice, purchase, progress=None):
        """
        Cognize the given invoice against the given SMS purchase.
        """
        unmatched_invoice = {}
        for row in invoice.data_rows:
            if row.product:
                unmatched_invoice.setdefault(row.product.upc, []).append(row)

        unmatched_purchase = {}
        for detail in purchase.details:
            upc = GPC(detail.F01, calc_check_digit='upc')
            unmatched_purchase.setdefault(upc, []).append(detail)

        for upc in list(unmatched_invoice.keys()):
            if upc in unmatched_purchase:

                # Match invoice rows to PO details.
                rows = unmatched_invoice[upc]
                details = unmatched_purchase[upc]
                matches, leftovers = self.match_rows(rows, details)
                if len(matches) + len(leftovers[0]) != len(rows):
                    raise RuntimeError("Had {0} invoice rows for {1} but {2} matches and {3} leftovers".format(
                            len(rows), upc, len(matches), len(leftovers[0])))
                if len(matches) + len(leftovers[1]) != len(details):
                    raise RuntimeError("Had {0} purchase detail lines for {1} but {2} matches and {3} leftovers".format(
                            len(details), upc, len(matches), len(leftovers[1])))

                # For matches, see if quantities etc. differ.
                for row, detail in matches:
                    row.line_number = detail.F1101
                    result = self.lines_differ(row, detail)
                    if result:
                        row.status_code = row.STATUS_DIFFERS_FROM_PURCHASE
                        row.status_text = result

                # For non-matches, handle accordingly.
                rows, details = leftovers
                for row in rows:
                    row.status_code = row.STATUS_NOT_IN_PURCHASE
                for detail in details:
                    self.add_purchase_detail(invoice, detail)

                # Remove "unmatched" references.
                del unmatched_invoice[upc]
                del unmatched_purchase[upc]

            else: # UPC in invoice but not PO.
                for row in unmatched_invoice[upc]:
                    row.status_code = row.STATUS_NOT_IN_PURCHASE
                del unmatched_invoice[upc]

        # If there's anything left in the SMS PO, then we must add dummy rows to
        # the invoice, to convey the situation to the user.
        assert len(unmatched_invoice) == 0
        for upc in list(unmatched_purchase.keys()):
            for detail in unmatched_purchase[upc]:
                self.add_purchase_detail(invoice, detail)

    def match_rows(self, rows, details):
        """
        For the given invoice rows and purchase details (all of which are
        assumed to contain the same UPC), figure out how they (don't) match
        up.  Returns a 2-tuple:

        The first element of the tuple returned is a list of "matches".
        Each element in this list is a 2-tuple of ``(invoice_row,
        purchase_detail)``.

        The second element of the tuple returned is another 2-tuple of
        "leftovers".  The first element in this is a list of invoice rows
        for which no PO details were found.  The second element is a list
        of PO details for which no invoice rows were found.
        """
        matches = []
        leftovers = {'rows': [], 'details': []}

        if len(rows) > len(details): # more rows than details

            rows = list(rows)
            for detail in details:
                found = False
                for i, row in enumerate(rows):
                    if row.ordered_cases == detail.F75 and row.ordered_units == detail.F1266:
                        matches.append(rows.pop(i), detail)
                        found = True
                        break
                if not found:
                    matches.append((rows.pop(), detail))
            leftovers['rows'].extend(rows)

        else: # at least as many details as rows

            details = list(details)
            for row in rows:
                if details:
                    found = False
                    for i, detail in enumerate(details):
                        if detail.F75 == row.ordered_cases and detail.F1266 == row.ordered_units:
                            matches.append(row, details.pop(i))
                            found = True
                            break
                    if not found:
                        matches.append((row, details.pop()))
                else:
                    leftovers['rows'].append(row)
            if details:
                leftovers['details'].extend(details)

        return matches, (leftovers['rows'], leftovers['details'])

    def add_purchase_detail(self, invoice, detail):
        """
        Adds a new row to the invoice, based on the given purchase detail.
        """
        upc = GPC(detail.F01, calc_check_digit='upc')
        product = get_product_by_upc(self.session, upc)
        return invoice.add_row(
            product=product,
            upc=upc,
            vendor_code=detail.F26,
            brand_name=product.brand.name if product and product.brand else None,
            description=product.description if product else detail.F1041,
            size=product.size if product else None,
            case_quantity=detail.F19,
            ordered_cases=detail.F75,
            ordered_units=detail.F1266,
            shipped_cases=0,
            shipped_units=0,
            case_cost=detail.F38,
            unit_cost=detail.F38 / int(detail.F19),
            total_cost=0,
            line_number=detail.F1101,
            status_code=VendorInvoiceRow.STATUS_NOT_IN_INVOICE)

    def lines_differ(self, invoice_row, purchase_detail):
        """
        Determine whether the values differ between an invoice row and a SMS
        purchase detail line.  If they are different, returns a string
        describing why; otherwise returns ``None``.
        """
        if invoice_row.shipped_cases:
            if not purchase_detail.F1003:
                return "invoice shows {0} cases shipped but PO shows none".format(
                    invoice_row.shipped_cases)
            if int(purchase_detail.F1003) != invoice_row.shipped_cases:
                return "invoice shows {0} cases shipped but PO shows {1}".format(
                    invoice_row.shipped_cases, repr(purchase_detail.F1003))
        elif purchase_detail.F1003:
            return "invoice shows no cases shipped but PO shows {0}".format(
                repr(purchase_detail.F1003))

        if invoice_row.shipped_units:
            if not purchase_detail.F70:
                return "invoice shows {0} units shipped but PO shows none".format(
                    invoice_row.shipped_units)
            if int(purchase_detail.F70) != invoice_row.shipped_units:
                return "invoice shows {0} units shipped but PO shows {1}".format(
                    invoice_row.shipped_units, repr(purchase_detail.F70))
        elif purchase_detail.F70:
            return "invoice shows no units shipped but PO shows {0}".format(
                repr(purchase_detail.F70))

        if invoice_row.case_cost != purchase_detail.F38:
            return "invoice shows case cost of {0} but PO shows {1}".format(
                repr(invoice_row.case_cost), repr(purchase_detail.F38))

        if invoice_row.total_cost != purchase_detail.F65:
            return "invoice shows total cost of {0} but PO shows {1}".format(
                repr(invoice_row.total_cost), repr(purchase_detail.F65))

    def get_execute_title(self, batch):
        """
        Get a human-friendly string describing the execution step for a batch.
        """
        if batch.purchase_order_number:
            return "Import as update to SMS Trans. No. {0}".format(batch.purchase_order_number)
        return "Import as new SMS PO"

    def execute(self, batch, progress=None):
        """
        Generate a SIL file which will either update an existing PO, or create
        a new one, and export the file to the local SMS node's inbox.
        """
        writer = SMSWriter()

        # Filter rows to exclude those which shouldn't be pushed to SMS.
        ignored = (VendorInvoiceRow.STATUS_NOT_IN_DB,
                   VendorInvoiceRow.STATUS_COST_NOT_IN_DB,
                   VendorInvoiceRow.STATUS_NOT_IN_INVOICE)
        rows = filter(lambda row: not row.removed
                      and row.status_code not in ignored,
                      batch.data_rows)
        if not rows:
            writer.close()
            os.remove(writer.sil_path)
            return True

        # Convert our invoice rows into purchase details.
        converted_details = [
            locsms.PurchaseDetail(
                F1101=r.line_number,
                F01=unicode(r.upc)[:-1],
                F1003=r.shipped_cases,
                F70=r.shipped_units,
                F19=r.case_quantity,
                F38=r.case_cost,
                F65=r.total_cost)
            for r in rows]

        if batch.purchase_order_number:

            # Fetch purchase from SMS database.
            purchase = self.get_purchase_order(batch.purchase_order_number)
            if not purchase:
                raise ValueError("Invalid SMS transaction number {0} for invoice: {1}".format(
                        repr(batch.purchase_order_number), repr(batch)))
                
            # Rewrite details according to our invoice.
            purchase.details = converted_details
            writer.write_receiving_update(purchase)

        else:
            # No PO number; create new PO from our invoice.
            purchase = locsms.Purchase(
                F1032=int(writer.consume_batch_id()),
                F27=batch.vendor.id,
                F334=batch.vendor.name,
                details=converted_details)
            writer.write_new_purchase(purchase)

        # Deploy SIL to SMS.
        writer.close()
        deploy_to_outbox(self.config, writer.sil_path)
        os.remove(writer.sil_path)
        return True
